using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using VertexPNT = Microsoft.DirectX.Direct3D.CustomVertex.PositionNormalTextured;
using VertexPCT = Microsoft.DirectX.Direct3D.CustomVertex.PositionColoredTextured;

namespace AMF
{
    public struct ExtendedTexture
    {
        public Texture texture;
        String _FileName;
        String _DirectoryName;
        String _FullPath;
        void DefaultFileNames()
        {
            if (_DirectoryName == null) _DirectoryName = "";
            if (_FullPath == null) _FullPath = "";
            if (_FileName == null) _FileName = "";
        }
        public String FileName
        {
            get
            {
                return _FileName;
            }
            set
            {
                try
                {
                    DefaultFileNames();
                    _FileName = Path.GetFileName(value);
                    if (_DirectoryName != "")
                    {
                        _FullPath = Path.Combine(_DirectoryName, value);
                    }
                    else
                    {
                        _FullPath = _FileName;
                    }
                }
                catch (Exception)
                {
                    DefaultFileNames();
                }
            }
        }
        public String DirectoryName
        {
            get
            {
                return _DirectoryName;
            }
            set
            {
                try
                {
                    DefaultFileNames();
                    if (_FileName != "")
                    {
                        _FullPath = Path.Combine(value, _FileName);
                        _DirectoryName = Path.GetFullPath(value);
                    }
                }
                catch (Exception)
                {
                    DefaultFileNames();
                }
            }
        }
        public String FullPath
        {
            get
            {
                return _FullPath;
            }
            set
            {
                try
                {
                    _FullPath = Path.GetFullPath(value);
                    _DirectoryName = Path.GetDirectoryName(value);
                    _FileName = Path.GetFileName(value);
                }
                catch (Exception)
                {
                    DefaultFileNames();
                }
            }
        }
        public Boolean Load(Device device, String folderName)
        {
            DirectoryName = folderName;
            return Load(device, true);
        }
        public Boolean Load(Device device, Boolean useFullPath)
        {
            try
            {
                // TextureLoader.FromStream(Device device, Stream stream, int readBytes);
                if (useFullPath) texture = TextureLoader.FromFile(device, FullPath);
                else texture = TextureLoader.FromFile(device, FileName);
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }
        public Boolean Save(Device device, String folderName)
        {
            DirectoryName = folderName;
            return Save(device, true);
        }
        public Boolean Save(Device device, Boolean useFullPath)
        {
            try
            {
                if (useFullPath) TextureLoader.Save(FullPath, ImageFileFormat.Png, texture);
                else TextureLoader.Save(FileName, ImageFileFormat.Png, texture);
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }
    }
    public class MeshLoader
    {
        // Configurations
        private PresentParameters presentParams;
        private Boolean isWindowed;
        public Boolean pauseRendering;
        private Control renderWindow;
        private Device device;
        // Mesh
        private String _name;
        public String name { get { return _name; } }
        private Mesh mesh;
        private ExtendedTexture[] textures;
        private Material[] materials;
        private Material blankMaterial;
        // Orientations
        private float objRadius;
        private float objScale;
        private Vector3 objCentre;
        private Quaternion objRotation;
        private float camPosition;
        // View options
        private Boolean _wireFrameMode;
        private Boolean _useLighting;
        public Boolean UseTextures;
        private Boolean _filterTextures;
        // Gex file mapping
        private GexFile importFile;
        // Test Results
        private DXFileRefType lastDXFileRefType;

        public enum CullMode
        {
            None = 1,
            Clockwise = 2,
            CounterClockwise = 3
        }

        public enum EffectType
        {
            Default = 1,
            SR1Object = 2,
            SR1Area = 3
        }

        public enum DXFileRefType
        {
            Absolute = 0,
            Relative = 1
        }

        public MeshLoader(Control renderWindow, Boolean isWindowed)
        {
            this.renderWindow = renderWindow;
            this.isWindowed = isWindowed;
            _wireFrameMode = false;
            _useLighting = true;
            UseTextures = true;
            _filterTextures = true;
            pauseRendering = false;
            lastDXFileRefType = DXFileRefType.Absolute;
            importFile = null;
            mesh = null;
            materials = new Material[1];
            textures = new ExtendedTexture[1];
            blankMaterial = new Material();
            blankMaterial.Diffuse = Color.White;
            CreateDevice();
            SetUpDevice();
            SetUpLighting();
            SetupCamera();
            renderWindow.Paint += new PaintEventHandler(Render);
        }

        private Boolean CreateDevice()
        {
            try
            {
                presentParams = new PresentParameters();
                presentParams.Windowed = isWindowed;
                presentParams.SwapEffect = SwapEffect.Discard;
                presentParams.EnableAutoDepthStencil = true;
                presentParams.AutoDepthStencilFormat = DepthFormat.D16;

                if (!isWindowed)
                {
                    presentParams.BackBufferFormat = Manager.Adapters[0].CurrentDisplayMode.Format;
                    presentParams.BackBufferCount = 1;
                    presentParams.BackBufferWidth = Manager.Adapters[0].CurrentDisplayMode.Width;
                    presentParams.BackBufferHeight = Manager.Adapters[0].CurrentDisplayMode.Height;
                    presentParams.FullScreenRefreshRateInHz = Manager.Adapters[0].CurrentDisplayMode.RefreshRate;
                    presentParams.PresentationInterval = PresentInterval.Immediate;
                }

                device = new Device(
                    0, DeviceType.Hardware, renderWindow,
                    CreateFlags.SoftwareVertexProcessing,
                    presentParams
                );
            }
            catch (DirectXException)
            {
                return false;
            }
            return true;
        }

        private void SetUpDevice()
        {
            device.RenderState.CullMode = Cull.None;
            device.RenderState.ZBufferEnable = true;
            device.RenderState.ZBufferWriteEnable = true;
            device.RenderState.DitherEnable = true;
            device.RenderState.SpecularEnable = false;
            device.RenderState.NormalizeNormals = true;
            device.RenderState.AlphaTestEnable = true;
            device.RenderState.AlphaFunction = Compare.Greater;
            device.RenderState.ReferenceAlpha = 0xA0;
            device.SetTextureStageState(
                0, TextureStageStates.ColorOperation, (int)TextureOperation.Modulate
            );
            device.SetTextureStageState(
                0, TextureStageStates.ColorArgument1, (int)TextureArgument.TextureColor
            );
            device.SetTextureStageState(
                0, TextureStageStates.ColorArgument2, (int)TextureArgument.Current
            );
            device.SetTextureStageState(
                0, TextureStageStates.AlphaOperation, (int)TextureOperation.Modulate
            );
            device.SetTextureStageState(
                0, TextureStageStates.AlphaArgument1, (int)TextureArgument.TextureColor
            );
            device.SetTextureStageState(
                0, TextureStageStates.AlphaArgument2, (int)TextureArgument.Diffuse
            );
            device.SetSamplerState(0, SamplerStageStates.MagFilter, (int)TextureFilter.Linear);
            device.SetSamplerState(0, SamplerStageStates.MinFilter, (int)TextureFilter.Linear);
            return;
        }

        private void SetUpLighting()
        {
            device.RenderState.Lighting = true;
            device.RenderState.Ambient =
                Color.FromArgb((int)0xFF, (int)0xFF, (int)0xFF, (int)0xFF);
            device.Lights[0].Type = LightType.Directional;
            device.Lights[0].Diffuse = Color.FromArgb((int)0xFF, (int)0xFF, (int)0xFF, (int)0xFF);
            device.Lights[0].Range = 1000f;
            device.Lights[0].Direction = new Vector3(0.0f, -0.3f, 0.5f);
            device.Lights[0].Direction.Normalize();
            device.Lights[0].Enabled = true;
            return;
        }

        public Boolean UseLighting
        {
            get
            {
                return _useLighting;
            }
            set
            {
                _useLighting = value;
                device.RenderState.Lighting = _useLighting;
                if (!_useLighting)
                {
                    device.SetTextureStageState(
                       0, TextureStageStates.ColorOperation, (int)TextureOperation.SelectArg1
                    );
                }
                else if (importFile != null && importFile.isArea)
                {
                    SetLightingMode(EffectType.SR1Area);
                }
                else
                {
                    SetLightingMode(EffectType.Default);
                }
            }
        }

        public void SetLightingMode(EffectType mode)
        {
            if (mode == EffectType.SR1Area)
            {
                device.RenderState.ColorVertex = true;
                device.RenderState.Lighting = false;
                device.RenderState.AmbientMaterialSource = ColorSource.Color1;
                device.SetTextureStageState(
                    0, TextureStageStates.ColorOperation, (int)TextureOperation.Modulate4X
                );
            }
            else
            {
                device.RenderState.ColorVertex = false;
                device.RenderState.AmbientMaterialSource = ColorSource.Material;
                device.SetTextureStageState(
                    0, TextureStageStates.ColorOperation, (int)TextureOperation.Modulate
                );
            }
            return;
        }

        private void SetupCamera()
        {
            objRadius = 1;
            objScale = 1;
            objCentre = new Vector3(0, 0, 0);
            device.Transform.Projection = Matrix.PerspectiveFovLH(
                (float)Math.PI / 4,
                (float)renderWindow.Width / (float)renderWindow.Height,
                0.1f, 100.0f
            );
            ResetTransforms();
            return;
        }

        public void ResetViewModes()
        {
            UseLighting = true;
            UseTextures = true;
            WireFrameMode = false;
            FilterTextures = true;
            SetCullMode(CullMode.None);
            ResetTransforms();
            return;
        }

        public void ResetTransforms()
        {
            objRotation = Quaternion.RotationYawPitchRoll(0, 0, 0);
            camPosition = -8.0f;
            objScale = 6 / (2 * objRadius);
            UpdateTransforms();
            return;
        }

        private void UpdateTransforms()
        {
            device.Transform.View = Matrix.LookAtLH(
                new Vector3(0, 0, camPosition),
                new Vector3(0, 0, camPosition + 8),
                new Vector3(0, 1, 0)
            );
            device.Transform.World =
            Matrix.Translation(-objCentre) *
            Matrix.Scaling(objScale, objScale, objScale) *
            Matrix.RotationYawPitchRoll(objRotation.X, objRotation.Y, objRotation.Z);
            return;
        }

        public void ZoomIn()
        {
            camPosition += 0.2f;
            UpdateTransforms();
            return;
        }

        public void ZoomOut()
        {
            camPosition -= 0.2f;
            UpdateTransforms();
            return;
        }

        public void RotateUp()
        {
            objRotation.Y += (float)(2 * Math.PI / 100);
            objRotation.Y %= (float)(2 * Math.PI);
            UpdateTransforms();
            return;
        }

        public void RotateDown()
        {
            objRotation.Y -= (float)(2 * Math.PI / 100);
            objRotation.Y %= (float)(2 * Math.PI);
            UpdateTransforms();
            return;
        }

        public void RotateLeft()
        {
            objRotation.X += (float)(2 * Math.PI / 100);
            objRotation.X %= (float)(2 * Math.PI);
            UpdateTransforms();
            return;
        }

        public void RotateRight()
        {
            objRotation.X -= (float)(2 * Math.PI / 100);
            objRotation.X %= (float)(2 * Math.PI);
            UpdateTransforms();
            return;
        }

        public void SetCullMode(CullMode mode)
        {
            device.RenderState.CullMode = (Cull)mode;
            return;
        }

        public Boolean WireFrameMode
        {
            get
            {
                return _wireFrameMode;
            }
            set
            {
                _wireFrameMode = value;
                if (_wireFrameMode)
                    device.RenderState.FillMode = FillMode.WireFrame;
                else
                    device.RenderState.FillMode = FillMode.Solid;
            }
        }

        private Mesh NormalizeMesh(Mesh mesh)
        {
            int[] adjacancy;
            mesh = mesh.Clone(
                MeshFlags.Managed, mesh.VertexFormat | VertexFormats.Normal,
                device
            );
            adjacancy = new int[mesh.NumberFaces * 3];
            mesh.GenerateAdjacency(0.01f, adjacancy);
            mesh.ComputeNormals(adjacancy);
            return mesh;
        }

        public void LoadDXMesh(String fileName)
        {
            ExtendedMaterial[] exMaterials;
            EffectInstance[] effects;
            try
            {
                if (fileName == null || fileName == "") return;
                mesh = Mesh.FromFile(
                    fileName, MeshFlags.Managed,
                    device, out exMaterials, out effects
                );
                mesh = NormalizeMesh(mesh);
            }
            catch (DirectXException) { return; }
            catch (AccessViolationException) { return; }

            if ((exMaterials != null) && (exMaterials.Length > 0))
            {
                materials = new Material[exMaterials.Length];
                textures = new ExtendedTexture[exMaterials.Length];
                for (int i = 0; i < exMaterials.Length; i++)
                {
                    materials[i] = exMaterials[i].Material3D;
                    textures[i].FullPath = exMaterials[i].TextureFilename;
                    textures[i].Load(device, true);
                }
            }
            else
            {
                materials = new Material[1];
                materials[0] = blankMaterial;
                textures = new ExtendedTexture[1];
            }
            GraphicsStream vertexData = mesh.VertexBuffer.Lock(0, 0, LockFlags.None);
            objRadius = Geometry.ComputeBoundingSphere(
                vertexData,
                mesh.NumberVertices,
                mesh.VertexFormat,
                out objCentre
            );
            mesh.VertexBuffer.Unlock();

            int extPos = fileName.LastIndexOf('.');
            if (extPos < 0) extPos = fileName.Length;
            int namePos = 1 + fileName.LastIndexOf('\\');
            _name = fileName.Substring(
                namePos,
                extPos - namePos
            );
            importFile = null;

            Boolean isSR1Room = false;
            for (int e =0; e < effects.Length; e++)
            {
                if (Path.GetFileNameWithoutExtension(effects[e].EffectFilename) ==
                    "SoulReaver1Room")
                {
                    isSR1Room = true;
                    break;
                }
            }
            if (isSR1Room) SetLightingMode(EffectType.SR1Area);
            else SetLightingMode(EffectType.Default);
            ResetViewModes();
            return;
        }

        public void SaveDXMesh(
            String fileName,
            XFileFormat xFormat,
            Boolean useEffects,
            Boolean retestExistingFiles
        )
        {
            // Check if ok to save
            if (mesh == null ||
                fileName == null ||
                fileName == "")
                return;

            // Generate the adjacency
            int[] adjacency = new int[mesh.NumberFaces * 3];
            mesh.GenerateAdjacency(0.01f, adjacency);

            // Get the materials
            ExtendedMaterial[] materials;

            // Save the file
            if (useEffects)
            {
                // Get the file and directory names
                String directory = Path.GetDirectoryName(fileName);

                // If the existing file does not own the folder,
                // append the new folder to the directory.
                if (retestExistingFiles) TestDXFileRefType(fileName);
                if (lastDXFileRefType == DXFileRefType.Absolute)
                {
                    directory = Path.ChangeExtension(fileName, null);
                    Directory.CreateDirectory(directory);
                }
                fileName = Path.Combine(directory, Path.GetFileName(fileName));

                // Get the materials
                materials = ExMaterials(ref textures, ref this.materials, true);

                // Save the fx file
                String effectFile = SaveDXEffect(directory, EffectMode);

                // Create fx file data
                EffectInstance[] effects = new EffectInstance[textures.Length];
                EffectDefault[] defaults;
                for (UInt16 e = 0; e < textures.Length; e++)
                {
                    if (EffectMode == EffectType.SR1Area)
                    {
                        defaults = new EffectDefault[1];
                    }
                    else
                    {
                        defaults = new EffectDefault[2];
                    }

                    // Store the texture names.  New textures have no alpha!
                    defaults[0] = new EffectDefault();
                    defaults[0].Type = EffectDefaultType.String;
                    defaults[0].ParameterName = "texture0";
                    defaults[0].Data = textures[e].FileName;

                    // If it's an area store the diffuse colours
                    if (EffectMode != EffectType.SR1Area)
                    {
                        float[] diffuse0 = {
                            this.materials[e].Diffuse.R / 255,
                            this.materials[e].Diffuse.G / 255,
                            this.materials[e].Diffuse.B / 255,
                            this.materials[e].Diffuse.A / 255
                        };
                        defaults[1] = new EffectDefault();
                        defaults[1].Type = EffectDefaultType.Floats;
                        defaults[1].ParameterName = "diffuse0";
                        defaults[1].Data = diffuse0;
                    }

                    // Create the effect instances
                    effects[e] = new EffectInstance();
                    effects[e].EffectFilename = Path.GetFileName(effectFile);
                    effects[e].SetDefaults(defaults);
                }

                // Save the textures
                SaveTexturesToFolder(directory);

                // Save x file with fx file data
                mesh.Save(fileName, adjacency, materials, effects, xFormat);
            }
            else
            {
                // Get the attributes
                materials = ExMaterials(ref textures, ref this.materials, false);

                // Save x file without x file data
                mesh.Save(fileName, adjacency, materials, xFormat);
            }
            return;
        }

        private DXFileRefType TestDXFileRefType(String fileName)
        {
            // Test the file for relative or absolute paths
            ExtendedMaterial[] exMaterials;
            EffectInstance[] effects;
            Mesh existingMesh = Mesh.FromFile(
                fileName, MeshFlags.Managed,
                device, out exMaterials, out effects
            );
            if (existingMesh == null) goto RetRelative;
            existingMesh.Dispose();
            for (int m = 0; m < exMaterials.Length; m++)
            {
                if (exMaterials[m].TextureFilename == null ||
                    exMaterials[m].TextureFilename == "" ||
                    exMaterials[m].TextureFilename == "(None)")
                {
                    continue;
                }
                if (Path.GetFileName(exMaterials[m].TextureFilename) ==
                    exMaterials[m].TextureFilename)
                {
                    goto RetRelative;   // Existing file owns the folder
                }
            }
            for (int e = 0; e < effects.Length; e++)
            {
                if (effects[e].EffectFilename == null ||
                    effects[e].EffectFilename == "" ||
                    effects[e].EffectFilename == "(None)")
                {
                    continue;
                }
                if (Path.GetFileName(effects[e].EffectFilename) ==
                    effects[e].EffectFilename)
                {
                    goto RetRelative;   // Existing file owns the folder
                }
            }

            lastDXFileRefType = DXFileRefType.Absolute;
            return lastDXFileRefType;

            RetRelative:
            lastDXFileRefType = DXFileRefType.Relative;
            return lastDXFileRefType;
        }

        public String SaveDXEffect(String directory, EffectType mode)
        {
            String fileName = null;
            switch (mode)
            {
                case EffectType.Default:
                    fileName = "SoulReaver1Object.fx";
                    break;
                case EffectType.SR1Object:
                    fileName = "SoulReaver1Object.fx";
                    break;
                case EffectType.SR1Area:
                    fileName = "SoulReaver1Room.fx";
                    break;
            }
            BinaryReader reader = new BinaryReader(
                System.Reflection.Assembly.GetExecutingAssembly().
                GetManifestResourceStream("ModelEx2." + fileName)
            );
            if (mode == EffectType.Default)
            {
                fileName = "ModelExDefault.fx";
            }
            fileName = Path.Combine(directory, fileName);
            FileStream fxFile = new FileStream(fileName, FileMode.Create);
            fxFile.Write(
                reader.ReadBytes((int)reader.BaseStream.Length),
                0, (int)reader.BaseStream.Length
            );
            reader.Close();
            fxFile.Close();

            return fileName;
        }

        public ExtendedMaterial[] ExMaterials(
            ref ExtendedTexture[] textures,
            ref Material[] materials,
            Boolean useLocalPath
        )
        {
            ExtendedMaterial[] MatArray = new ExtendedMaterial[textures.Length];
            for (int t = 0; t < textures.Length; t++)
            {
                MatArray[t] = new ExtendedMaterial();
                if (useLocalPath)
                {
                    MatArray[t].TextureFilename = textures[t].FileName;
                }
                else
                {
                    MatArray[t].TextureFilename = textures[t].FullPath;
                }
                MatArray[t].Material3D = materials[t];
            }
            return MatArray;
        }

        public void LoadGexMesh(String fileName)
        {
            if (fileName == null | fileName == "") return;

            try
            {
                importFile = new GexFile(fileName);
            }
            catch (Exception e)
            {
                throw e;
            }

            // Choose the vertex format
            VertexFormats vertexFormat;
            if (importFile.isObject) vertexFormat = VertexPNT.Format;
            else vertexFormat = VertexPCT.Format;

            // Create the mesh
            mesh = new Mesh(
                (int)importFile.polygonCount,
                (int)importFile.indexCount,
                MeshFlags.Managed,
                vertexFormat,
                device
            );

            // Insert the vertices
            if (vertexFormat == VertexPNT.Format)
            {
                VertexPNT[] verticesPNT = new VertexPNT[importFile.indexCount];
                for (UInt16 v = 0; v < importFile.indexCount; v++)
                {
                    verticesPNT[v] = new VertexPNT(
                        100f * (float)importFile.vertices[v].worldPos.x,
                        100f * (float)importFile.vertices[v].worldPos.z,
                        100f * (float)importFile.vertices[v].worldPos.y,
                        (float)importFile.vertices[v].normal.x,
                        (float)importFile.vertices[v].normal.z,
                        (float)importFile.vertices[v].normal.y,
                        ((float)(importFile.vertices[v].u) / 255f) + (0.5f / 255),
                        ((float)(importFile.vertices[v].v) / 255f) + (0.5f / 255)
                    );
                }
                mesh.VertexBuffer.SetData(verticesPNT, 0, LockFlags.None);
            }
            else
            {
                VertexPCT[] verticesPCT = new VertexPCT[importFile.indexCount];
                for (UInt16 v = 0; v < importFile.indexCount; v++)
                {
                    verticesPCT[v] = new VertexPCT(
                        100f * (float)importFile.vertices[v].worldPos.x,
                        100f * (float)importFile.vertices[v].worldPos.z,
                        100f * (float)importFile.vertices[v].worldPos.y,
                        (int)importFile.vertices[v].colour,
                        ((float)(importFile.vertices[v].u) / 255f) + (0.5f / 255),
                        ((float)(importFile.vertices[v].v) / 255f) + (0.5f / 255)
                    );
                }
                mesh.VertexBuffer.SetData(verticesPCT, 0, LockFlags.None);
            }

            // Insert the indices
            UInt16[] indices = new UInt16[importFile.indexCount];
            for (UInt16 i = 0; i < importFile.indexCount; i++)
            {
                if (importFile.indices[i] % 3 == 0) indices[i] = importFile.indices[i];
                if (importFile.indices[i] % 3 == 1) indices[i] = importFile.indices[i + 1];
                if (importFile.indices[i] % 3 == 2) indices[i] = importFile.indices[i - 1];
            }
            mesh.IndexBuffer.SetData(indices, 0, LockFlags.None);

            // Load the materials
            Material newMaterial = new Material();
            materials = new Material[importFile.materialCount];
            textures = new ExtendedTexture[importFile.materialCount];
            AttributeRange[] attributeRange = new AttributeRange[importFile.materialCount];
            for (UInt16 m = 0; m < importFile.materialCount; m++)
            {
                if (importFile.isObject && importFile.materials[m].textureUsed)
                {
                    newMaterial.Diffuse = Color.White;
                }
                else
                {
                    newMaterial.Diffuse = Color.FromArgb(
                        (Int32)(importFile.materials[m].colour)
                    );
                }
                materials[m] = newMaterial;
                textures[m].FileName = importFile.materials[m].textureName;
                attributeRange[m].AttributeId = m;
            }

            // Get the attributes
            Int32[] attributeArray = mesh.LockAttributeBufferArray(LockFlags.Discard);
            for (UInt16 p = 0; p < importFile.polygonCount; p++)
            {
                attributeArray[p] = importFile.polygons[p].material.ID;
            }
            mesh.UnlockAttributeBuffer(attributeArray);

            // Set the lighting mode
            if (importFile.isArea)
                SetLightingMode(EffectType.SR1Area);
            else
                SetLightingMode(EffectType.Default);

            // Weld any duplicate vertices
            mesh.WeldVertices(
                WeldEpsilonsFlags.WeldPartialMatches,
                new WeldEpsilons(),
                null
            );

            // If it's an area, compute the normals
            if (importFile.isArea)
            {
                mesh = mesh.Clone(
                    MeshFlags.Managed, mesh.VertexFormat | VertexFormats.Normal,
                    device
                );
                mesh = NormalizeMesh(mesh);
            }

            // Centre the mesh
            GraphicsStream vertexData = mesh.VertexBuffer.Lock(0, 0, LockFlags.None);
            objRadius = Geometry.ComputeBoundingSphere(
                vertexData,
                mesh.NumberVertices,
                mesh.VertexFormat,
                out objCentre
            );
            mesh.VertexBuffer.Unlock();

            _name = importFile.modelName.TrimEnd(new char[] { '_' });
            ResetViewModes();
            return;
        }

        public Boolean LoadTexture(String fileName, int index)
        {
            textures[index].FullPath = fileName;
            return textures[index].Load(device, true);
        }

        public Boolean LoadTexturesFromFolder(String folderName)
        {
            if (mesh == null) return false;
            for (UInt16 t = 0; t < textures.Length; t++)
            {
                textures[t].Load(device, folderName);
            }
            return true;
        }

        public Boolean SaveTexturesToFolder(String folderName)
        {
            if (mesh == null) return false;
            for (UInt16 t = 0; t < textures.Length; t++)
            {
                textures[t].Save(device, folderName);
            }
            return true;
        }

        public String[] TextureNames
        {
            get
            {
                string[] fileNames = new string[textures.Length];
                for (int i = 0; i < textures.Length; i++)
                {
                    fileNames[i] = textures[i].FileName;
                    if (fileNames[i] == null || fileNames[i] == "")
                    {
                        fileNames[i] = "(None)";
                    }
                }
                return fileNames;
            }
        }

        public String FXFileName
        {
            get
            {
                switch (EffectMode)
                {
                    case EffectType.SR1Object:
                        return "SoulReaver1Object.fx";
                    case EffectType.SR1Area:
                        return "SoulReaver1Room.fx";
                    default:
                        return "ModelExDefault.fx";
                }
            }
        }

        public String[] FindExistingFiles(String meshFileName)
        {
            // Break path into file and directories
            String[] fileNames = new String[textures.Length + 2];
            String file = Path.GetFileName(meshFileName);
            String directory = Path.GetDirectoryName(meshFileName);

            // If the existing file does not own the folder,
            // append the new folder to the directory.
            if (TestDXFileRefType(meshFileName) == DXFileRefType.Absolute)
            {
                directory = Path.Combine(directory, Path.GetFileNameWithoutExtension(meshFileName));
            }

            // Check mesh file
            fileNames[0] = Path.Combine(directory, file);
            if (!File.Exists(fileNames[0])) fileNames[0] = null;

            // Check effect file
            fileNames[1] = Path.Combine(directory, FXFileName);
            if (!File.Exists(fileNames[1])) fileNames[1] = null;

            // Check texture files
            for (int i = 0; i < textures.Length; i++)
            {
                if (textures[i].FileName == null || textures[i].FileName == "")
                {
                    fileNames[i+2] = null;
                    continue;
                }
                fileNames[i+2] = Path.Combine(directory, textures[i].FileName);
                if (!File.Exists(fileNames[i+2])) fileNames[i+2] = null;
            }
            return fileNames;
        }

        public EffectType EffectMode
        {
            get
            {
                if (importFile == null) return EffectType.Default;
                if (importFile.isObject) return EffectType.SR1Object;
                if (importFile.isArea) return EffectType.SR1Area;
                return EffectType.Default;
            }
        }

        public Boolean FilterTextures
        {
            get
            {
                return _filterTextures;
            }
            set
            {
                _filterTextures = value;
                if (_filterTextures)
                {
                    device.SetSamplerState(0, SamplerStageStates.MagFilter, (int)TextureFilter.Linear);
                    device.SetSamplerState(0, SamplerStageStates.MinFilter, (int)TextureFilter.Linear);
                }
                else
                {
                    device.SetSamplerState(0, SamplerStageStates.MagFilter, (int)TextureFilter.Point);
                    device.SetSamplerState(0, SamplerStageStates.MinFilter, (int)TextureFilter.Point);
                }
            }
        }

        public void Render(Object sender, PaintEventArgs e)
        {
            if (pauseRendering) return;
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Gray, 1.0f, 0);
            device.BeginScene();
            if (mesh != null)
            {
                for (int i = 0; i < textures.Length; i++)
                {
                    if (UseTextures)
                    {
                        device.Material = materials[i];
                        device.SetTexture(0, textures[i].texture);
                    }
                    else
                    {
                        device.Material = blankMaterial;
                        device.SetTexture(0, null);
                    }
                    mesh.DrawSubset(i);
                }
            }
            device.EndScene();
            device.Present();

            renderWindow.Invalidate();
            return;
        }
    }
}